from typing import List

import transaction
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import Session

from workspaces.capability.enums import CapabilityRequestState, ExecutionState
from workspaces.capability.helpers import CapabilitySequence
from workspaces.capability.helpers_interfaces import ParameterIF
from workspaces.capability.schema import (
    Capability,
    CapabilityExecution,
    CapabilityRequest,
    CapabilityVersion,
)
from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.interfaces import CapabilityInfoIF
from workspaces.products.schema_interfaces import FutureProductIF

Base = declarative_base()


class CapabilityName(object):
    pass


class CapabilityInfo(CapabilityInfoIF):
    """
    Interface for data access object that can look up and record information about capabilities
    and capability requests. Accesses and modifies the database.
    """

    def __init__(self, session: Session):
        self.session = session

    def create_capability(
        self, name: CapabilityName, steps: CapabilitySequence, max_jobs: int
    ) -> Capability:
        """
        Create new capability and save it in the database

        :param name: Name of new capability
        :param steps: List of capability steps required to execute the capability
        :param max_jobs: Max allowed number of concurrent executions of this kind of capability
        :return: Created capability
        """
        capability = Capability(name=name, steps=str(steps), max_jobs=max_jobs)
        self.save_entity(capability)
        return capability

    def create_capability_request(
        self,
        capability_name: str,
        parameters: List[ParameterIF] = None,
        future_products: List[FutureProductIF] = None,
    ) -> CapabilityRequest:
        """
        Create new capability request and save it in the database

        :param capability_name: name of the requested capability
        :param parameters: List
        :param future_products:
        :return: Created CapabilityRequest
        """
        capability = self.lookup_capability(capability_name)
        request = CapabilityRequest(
            state=CapabilityRequestState.Ready.name,
            capability=capability,
            parameters=str(parameters),
            # a trick here is to ensure that we always have a first version, with the original parameters
            versions=[CapabilityVersion(version_number=1, parameters=str(parameters))],
        )
        self.save_entity(request)
        return request

    def create_execution(self, request: CapabilityRequest) -> CapabilityExecution:
        """
        Create new execution record for a request and save it in the database

        :param request: ID of the capability request
        :return: Created CapabilityExecution
        """
        # look up the most recent version
        most_recent_version = request.versions[-1]

        execution = CapabilityExecution(
            state=ExecutionState.Ready.name,
            version=most_recent_version,
            current_step=0,
            # ensure that we have a copy of the step sequence as it was when the execution started
            steps=request.capability.steps,
        )
        self.save_entity(execution)
        return execution

    def lookup_capability(self, capability_name: str) -> Capability:
        return self.session.query(Capability).filter_by(name=capability_name).first()

    def lookup_capability_request(self, request_id) -> CapabilityRequest:
        return self.session.query(CapabilityRequest).filter_by(id=request_id).first()

    def lookup_execution(self, execution_id: int) -> CapabilityExecution:
        return (
            self.session.query(CapabilityExecution).filter_by(id=execution_id).first()
        )

    def save_entity(self, entity: Base):
        """
        Save entity to database

        :param entity:
        """
        self.session.add(entity)
        self.session.flush()

    def lookup_execution_by_workflow_request_id(
        self, workflow_request_id: int
    ) -> CapabilityExecution:
        """
        Find execution record given a capability request

        :param workflow_request_id:
        :return: Capability execution record
        """
        return (
            self.session.query(CapabilityExecution)
            .filter_by(workflow_request_id=workflow_request_id)
            .one()
        )

    def requests_for_capability(self, capability_name: str) -> List[CapabilityRequest]:
        """
        Gets all capability requests for a given capability

        :param capability_name: Name of capability
        :return: List of associated requests
        """
        capability = self.lookup_capability(capability_name)
        return (
            self.session.query(CapabilityRequest)
            .filter_by(capability_id=capability.id)
            .all()
        )

    def save_execution(self, execution: CapabilityExecutionIF):
        self.session.add(execution)
        transaction.commit()
