import threading
import time
from collections import namedtuple
from queue import Empty, PriorityQueue, Queue
from typing import NamedTuple, Optional

from workspaces.capability.schema_interfaces import CapabilityExecutionIF
from workspaces.capability.services.capability_engine import CapabilityEngine
from workspaces.capability.services.interfaces import (
    CapabilityInfoIF,
    CapabilityQueueIF,
)
from workspaces.workflow.services.interfaces import WorkflowServiceIF


class CapabilityQueue(CapabilityQueueIF):
    """
    Organizes capability executions in a priority order and makes it possible to control
    the number of concurrent executions or pause execution altogether
    """

    EngineList = namedtuple("EngineList", ["available", "in_use"])

    def __init__(
        self,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
        max_concurrency: int,
    ):
        self.capability_info = capability_info
        self.queue = PriorityQueue()
        self.max_concurrency = max_concurrency
        self.engine_list = self.init_engines(
            max_concurrency, capability_info, workflow_service
        )
        self.queue_manager = threading.Thread(
            target=self.process_executions, daemon=True
        )
        self.queue_manager.start()
        self.paused = False

    @classmethod
    def init_engines(
        cls,
        num_engines: int,
        capability_info: CapabilityInfoIF,
        workflow_service: WorkflowServiceIF,
    ) -> NamedTuple:
        """
        Initialize available engine queue and list of in-use engines

        :return: EngineList
        """
        engine_list = cls.EngineList(Queue(), [])
        for _ in range(num_engines):
            engine_list.available.put(
                CapabilityEngine(capability_info, workflow_service)
            )

        return engine_list

    def process_executions(self):
        """
        Process execution at the front of the queue; I imagine this would be called repeatedly
        with a small delay
        """
        while True:
            if self.queue.empty() is False:
                execution = self.queue.get()[1]
                engine = self.get_available_engine()
                if engine:
                    execution = self.capability_info.lookup_execution(execution.id)
                    # Start engine
                    engine.execute(execution)
                    # Move engine to in-use list
                    self.engine_list.in_use.append(engine)
                else:
                    # FIXME: Logging
                    print("No available engines. Try again later.")
            time.sleep(0.5)

    def find_engine(self, execution_id: int) -> CapabilityEngine:
        """
        Find in-use engine executing execution with given ID

        :param execution_id: ID of requested execution
        :return: Corresponding engine
        """
        for engine in self.engine_list.in_use:
            if engine.execution_id == execution_id:
                # Found correct engine
                return engine

    def complete_execution(self, execution_id: int):
        """
        Remove execution from its engine and return engine to available queue

        :param execution_id: ID of requested execution
        """
        engine = self.find_engine(execution_id)
        self.engine_list.in_use.remove(engine)
        engine.execution_id = None
        engine.execution = None
        self.engine_list.available.put(engine)

    def get_available_engine(self) -> Optional[CapabilityEngine]:
        """
        Get an engine if one is available

        :return: Engine if one is available, else False
        """
        try:
            return self.engine_list.available.get_nowait()
        except Empty:
            return None

    def enqueue(self, request: CapabilityExecutionIF, priority: int):
        self.queue.put((priority, request))

    def dequeue(self) -> CapabilityExecutionIF:
        return self.queue.get()

    def pause(self):
        # TODO: Implement
        self.paused = True

    def unpause(self):
        # TODO: Implement
        self.paused = False

    def __len__(self):
        return self.queue.qsize()
