""" Utilities and helper classes for PackageTester """
import os
import re
import subprocess
import sys
from enum import Enum
from pathlib import Path
from typing import List

TIMESTAMP_FORMAT = "%Y-%m-%dT%H-%M-%S"

PACKAGE_TESTER = "package_tester.py"
ENTRYPOINT = "package-tester"

# pylint: disable=R0903
sys.path.insert(0, str(Path("../src").absolute()))


def execute_shell_command(command: str) -> List[str]:
    """
    Run command in shell and return its output.

    :param command:
    :return: lines of command output
    """

    output_lines = []

    with subprocess.Popen(
        command,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        bufsize=1,
        universal_newlines=True,
        shell=True,
    ) as proc:

        for line in proc.stdout.readlines():
            line = line.strip()
            if len(line) > 0:
                output_lines.append(line)

    return output_lines


def get_project_root() -> Path:
    """
    Get the root of this project, TODO: without depending on name of repo
    or of Docker container top level.

    :return:
    """
    my_path = Path(__file__)
    path = my_path
    while not path.name.endswith("workspaces") and not path.name.endswith("code"):
        path = path.parent

    return path


def running_in_docker() -> bool:
    """
    Are we executing in shell (rather than in a Docker container or IDE)?

    :return:
    """
    return "SHELL" not in os.environ.keys()


class TestedPackageInfo:
    """ Holds metadata about a package's test(s) """

    def __init__(self, package_name: str, location: Path):

        self.package_name = package_name
        self.location = location
        self.summary = None

        self.test_output = []


class ResultType(Enum):
    """ These are statuses that pytest reports. """

    # The success string string may be "passed". Thanks, pytest.
    PASSES = "passes"
    PASSED = "passed"

    # Failure string may be "failed" or "failures". Thanks again.
    FAILED = "failed"
    FAILURES = "failures"

    SKIP = "skipped"
    ERROR = "error"


class PytestResult:
    """ Encapsulates final line of pytest output, which has the results. """

    def __init__(self, **kwargs):
        self.failures = (
            kwargs.get(ResultType.FAILED.value)
            or kwargs.get(ResultType.FAILURES.value)
            or 0
        )
        self.skips = kwargs.get(ResultType.SKIP.value) or 0
        self.errors = kwargs.get(ResultType.ERROR.value) or 0
        self.elapsed_time = kwargs.get("elapsed_time") or ""
        self.passes = (
            kwargs.get(ResultType.PASSED.value)
            or kwargs.get(ResultType.PASSES.value)
            or 0
        )

    def __str__(self):
        return (
            f"{self.passes} passes, "
            f"{self.errors} errors, "
            f"{self.failures} failures, "
            f"{self.skips} skipped; elapsed time: {self.elapsed_time}"
        )

    def __eq__(self, other):
        if not isinstance(other, PytestResult):
            return False

        if (
            other.passes == self.passes
            and other.failures == self.failures
            and other.errors == self.errors
            and other.skips == self.skips
        ):
            self_elapsed = self.elapsed_time
            other_elapsed = other.elapsed_time

            # We don't expect an exact match
            if self_elapsed is None:
                return other_elapsed is None
            return other_elapsed is not None

        return False

    def elapsed_time_numeric(self) -> float:
        """
        Grab numeric portion of "elapsed time" string

        :return:
        """
        match = re.search("([0-9.]+)", self.elapsed_time)
        return float(match.group(1)) if match else -1.0

    def is_empty(self) -> bool:
        """
        Did we actually get results for this package?

        :return:
        """
        return self.failures == self.passes == self.skips == self.errors == 0


class PytestResultsParser:
    """ Trawls final line of pytest output to get the meat. """

    def __init__(self, pytest_results_line: str):
        self.pytest_results_line = pytest_results_line

    def parse(self) -> PytestResult:
        """
        Get the 411 out of the pytest summary line.

        :return:
        """
        counts = {}
        for result_type in ResultType:
            count = self.get_count(result_type)
            if count < 0:
                count = 0
            counts[result_type] = count

        result = PytestResult(
            passed=counts[ResultType.PASSED] + counts[ResultType.PASSES],
            failed=counts[ResultType.FAILED] + counts[ResultType.FAILURES],
            skipped=counts[ResultType.SKIP],
            error=counts[ResultType.ERROR],
            elapsed_time=self.get_elapsed_time(),
        )
        return result

    def get_count(self, result_type: ResultType) -> int:
        """
        Find the count that pytest has reported for this result type.

        :param result_type:
        :return:
        """
        key = result_type.value
        if key not in self.pytest_results_line:
            return 0

        match = re.search(f"([0-9]+) {key}", self.pytest_results_line)
        if match:
            return int(match.group(1))
        return -1

    def get_elapsed_time(self) -> str:
        """
        Parse elapsed time out of pytest result line.

        :return:
        """
        match = re.search(r"in ([.0-9]+[a-z])", self.pytest_results_line)
        return match.group(1) if match else "-1.0"
